/*******************************************************************************
 * Copyright (c) 2007, 2018 Association for Decentralized Information Management
 * in Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.diagram.adapter;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.simantics.databoard.Bindings;
import org.simantics.db.ReadGraph;
import org.simantics.db.RequestProcessor;
import org.simantics.db.Resource;
import org.simantics.db.common.primitiverequest.OrderedSet;
import org.simantics.db.common.procedure.wrapper.SetListenerToSingleSetListener;
import org.simantics.db.common.request.BinaryRead;
import org.simantics.db.exception.DatabaseException;
import org.simantics.db.procedure.SetListener;
import org.simantics.diagram.stubs.DiagramResource;
import org.simantics.layer0.Layer0;
import org.simantics.scenegraph.profile.Group;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Tuukka Lehtonen
 */
public class TypeGroup implements Group {

    private static final Logger LOGGER = LoggerFactory.getLogger(TypeGroup.class);

    private final Collection<Resource> types;
    private final String               name;

    public TypeGroup(String name, Resource type) {
        this(name, new Resource[] { type });
    }

    public TypeGroup(String name, Resource... types) {
        this.types = Arrays.asList(types);
        this.name = name;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + types.hashCode();
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        TypeGroup other = (TypeGroup) obj;
        return types.equals(other.types);
    }

    /**
     * Initialize TypeGroup from a DIAGRAM.Group instance.
     * 
     * @param graph
     * @param group
     * @throws DatabaseException
     */
    public TypeGroup(ReadGraph graph, final Resource group) throws DatabaseException {
        Layer0 L0 = Layer0.getInstance(graph);
        DiagramResource DIA = DiagramResource.getInstance(graph);

        this.name = graph.getPossibleRelatedValue(group, L0.HasName, Bindings.STRING);
        this.types = graph.getObjects(group, DIA.TypeGroup_HasType);
    }

    @Override
    public void trackItems(RequestProcessor processor, final Resource runtimeDiagram, final SetListener<Resource> listener) throws DatabaseException {
        if (types.isEmpty()) {
            System.out.println("TypeGroup has no types!");
            return;
        }

        processor.syncRequest(new BinaryRead<Resource, Collection<Resource>, Collection<Resource>>(runtimeDiagram, types) {

            @Override
            public Set<Resource> perform(ReadGraph graph) throws DatabaseException {
                HashSet<Resource> result = new HashSet<Resource>();

                Resource realDiagram = graph.getPossibleObject(runtimeDiagram, DiagramResource.getInstance(graph).RuntimeDiagram_HasConfiguration);
                if (realDiagram == null)
                    return result;

//                System.out.println("looking for diagram elements of type:");
//                for (Resource t : types)
//                    System.out.println("\t" + NameUtils.getSafeName(graph, t, true));

                if (graph.hasStatement(realDiagram)) {
                    Collection<Resource> elements = graph.syncRequest(new OrderedSet(realDiagram));
                    for (Resource element : elements) {
    //                    System.out.println("checking element " + NameUtils.getSafeName(graph, element, true));
                        Collection<Resource> elementTypes = graph.getTypes(element);
                        if (!Collections.disjoint(types, elementTypes))
                            result.add(element);
                    }
                } else {
                    LOGGER.warn("Most likely after deleting a diagram or something therefore no ordered set can be found for {}", realDiagram);
                }
                return result;
            }

        }, new SetListenerToSingleSetListener<Resource>(listener));
    }

    @Override
    public String toString() {
        return "every '" + name + "'";
    }

}
